// Copyright © 2023 ab180. All rights reserved.

#include "AirbridgeAndroidCore.h"
#include "AirbridgeCallbacks.h"
#include "UObject/UObjectIterator.h"
#include "Misc/Guid.h"

#include "Android/AndroidApplication.h"
#include "Android/AndroidJNI.h"
#include <android/log.h>

FString AirbridgeAndroidCore::StartDeeplink;
TMap<FString, FString> AirbridgeAndroidCore::ReceivedAttribution;

AirbridgeAndroidCore::AirbridgeAndroidCore()
{
}

AirbridgeAndroidCore::~AirbridgeAndroidCore()
{
}

void AirbridgeAndroidCore::SetAirbridgeStringParamFunctionMap(
	const AirbridgeFunctionKey Key,
	const FString& UUID,
	const AirbridgeStringParamFunction& Function
)
{
	if (!Function) return;
	
	if (!AirbridgeStringParamFunctionMap.Contains(Key))
	{
		TMap<FString, AirbridgeStringParamFunction> Value;
		Value.Add(UUID, Function);
		
		AirbridgeStringParamFunctionMap.Add(Key, Value);
	}
	else
	{
		AirbridgeStringParamFunctionMap[Key].Add(UUID, Function);
	}
}

void AirbridgeAndroidCore::SetAirbridgeVoidParamFunctionMap(
	const AirbridgeFunctionKey Key,
	const FString& UUID,
	const AirbridgeVoidParamFunction& Function
)
{
	if (!Function) return;
	
	if (!AirbridgeVoidParamFunctionMap.Contains(Key))
	{
		TMap<FString, AirbridgeVoidParamFunction> Value;
		Value.Add(UUID, Function);
		
		AirbridgeVoidParamFunctionMap.Add(Key, Value);
	}
	else
	{
		AirbridgeVoidParamFunctionMap[Key].Add(UUID, Function);
	}
}

#pragma region Core

void AirbridgeAndroidCore::EnableSDK()
{
#if PLATFORM_ANDROID
	AirbridgeJNI::GameActivityCallMethod("airbridgeEnableSDK");
#endif
}

void AirbridgeAndroidCore::DisableSDK()
{
#if PLATFORM_ANDROID
	AirbridgeJNI::GameActivityCallMethod("airbridgeDisableSDK");
#endif
}

bool AirbridgeAndroidCore::IsSDKEnabled()
{
#if PLATFORM_ANDROID
	return AirbridgeJNI::GameActivityCallMethod_Return_Boolean("airbridgeIsSDKEnabled");
#endif
	return false;
}

#pragma endregion

#pragma region Privacy

void AirbridgeAndroidCore::StartTracking()
{
#if PLATFORM_ANDROID
	AirbridgeJNI::GameActivityCallMethod("airbridgeStartTracking");
#endif
}

void AirbridgeAndroidCore::StopTracking()
{
#if PLATFORM_ANDROID
	AirbridgeJNI::GameActivityCallMethod("airbridgeStopTracking");
#endif
}

bool AirbridgeAndroidCore::IsTrackingEnabled()
{
#if PLATFORM_ANDROID
	return AirbridgeJNI::GameActivityCallMethod_Return_Boolean("airbridgeIsTrackingEnabled");
#endif
	return false;
}

#pragma endregion

#pragma region Deeplink

void AirbridgeAndroidCore::SetOnDeeplinkReceived(const AirbridgeStringParamFunction OnDeeplinkReceived, bool UseBroadcast)
{
	// Only the first callback is valid
	if (DidSetUnrealBroadcastDeeplink || AirbridgeStringParamFunctionMap.Contains(AirbridgeFunctionKey::OnDeeplinkReceived))
	{
		return;
	}

	if (UseBroadcast)
	{
		DidSetUnrealBroadcastDeeplink = true;
	}
	else
	{
		TMap<FString, AirbridgeStringParamFunction> OnDeeplinkReceivedMap;
		OnDeeplinkReceivedMap.Add("", OnDeeplinkReceived);
	
		AirbridgeStringParamFunctionMap.Add(AirbridgeFunctionKey::OnDeeplinkReceived, OnDeeplinkReceivedMap);		
	}
	
	// Success
	if (!StartDeeplink.IsEmpty())
	{
		HandleDeeplinkOnSuccess(StartDeeplink);
	}
	// Check deferred deeplink
	else
	{
#if PLATFORM_ANDROID
		AirbridgeJNI::GameActivityCallMethod("airbridgeHandleDeferredDeeplink");
#endif
	}
}

#pragma endregion

#pragma region Data Collection

void AirbridgeAndroidCore::SetUserID(const FString& Id)
{
#if PLATFORM_ANDROID
	AirbridgeJNI::GameActivityCallMethod_OneParam_String("airbridgeSetUserID", Id);
#endif
}

void AirbridgeAndroidCore::ClearUserID()
{
#if PLATFORM_ANDROID
	AirbridgeJNI::GameActivityCallMethod("airbridgeClearUserID");
#endif
}

void AirbridgeAndroidCore::SetUserEmail(const FString& Email)
{
#if PLATFORM_ANDROID
	AirbridgeJNI::GameActivityCallMethod_OneParam_String("airbridgeSetUserEmail", Email);
#endif
}

void AirbridgeAndroidCore::ClearUserEmail()
{
#if PLATFORM_ANDROID
	AirbridgeJNI::GameActivityCallMethod("airbridgeClearUserEmail");
#endif
}

void AirbridgeAndroidCore::SetUserPhone(const FString& Phone)
{
#if PLATFORM_ANDROID
	AirbridgeJNI::GameActivityCallMethod_OneParam_String("airbridgeSetUserPhone", Phone);
#endif
}

void AirbridgeAndroidCore::ClearUserPhone()
{
#if PLATFORM_ANDROID
	AirbridgeJNI::GameActivityCallMethod("airbridgeClearUserPhone");
#endif
}

void AirbridgeAndroidCore::SetUserAttribute(const FString& Key, const int32 Value)
{
#if PLATFORM_ANDROID
    if (JNIEnv* Env = FAndroidApplication::GetJavaEnv())
    {
        jmethodID Method = FJavaWrapper::FindMethod(Env,
            FJavaWrapper::GameActivityClassID,
            "airbridgeSetUserAttribute", "(Ljava/lang/String;I)V", false);
        jstring jKey = Env->NewStringUTF(TCHAR_TO_UTF8(*Key));
        FJavaWrapper::CallVoidMethod(Env, FJavaWrapper::GameActivityThis, Method, jKey, Value);
    }
#endif
}

void AirbridgeAndroidCore::SetUserAttribute(const FString& Key, const int64 Value)
{
#if PLATFORM_ANDROID
    if (JNIEnv* Env = FAndroidApplication::GetJavaEnv())
    {
        jmethodID Method = FJavaWrapper::FindMethod(Env,
            FJavaWrapper::GameActivityClassID,
            "airbridgeSetUserAttribute", "(Ljava/lang/String;J)V", false);
        jstring jKey = Env->NewStringUTF(TCHAR_TO_UTF8(*Key));
        FJavaWrapper::CallVoidMethod(Env, FJavaWrapper::GameActivityThis, Method, jKey, Value);
    }
#endif
}

void AirbridgeAndroidCore::SetUserAttribute(const FString& Key, const float Value)
{
#if PLATFORM_ANDROID
    if (JNIEnv* Env = FAndroidApplication::GetJavaEnv())
    {
        jmethodID Method = FJavaWrapper::FindMethod(Env,
            FJavaWrapper::GameActivityClassID,
            "airbridgeSetUserAttribute", "(Ljava/lang/String;F)V", false);
        jstring jKey = Env->NewStringUTF(TCHAR_TO_UTF8(*Key));
        FJavaWrapper::CallVoidMethod(Env, FJavaWrapper::GameActivityThis, Method, jKey, Value);
    }
#endif
}

void AirbridgeAndroidCore::SetUserAttribute(const FString& Key, const double Value)
{
#if PLATFORM_ANDROID
	if (JNIEnv* Env = FAndroidApplication::GetJavaEnv())
	{
		jmethodID Method = FJavaWrapper::FindMethod(Env,
			FJavaWrapper::GameActivityClassID,
			"airbridgeSetUserAttribute", "(Ljava/lang/String;D)V", false);
		jstring jKey = Env->NewStringUTF(TCHAR_TO_UTF8(*Key));
		FJavaWrapper::CallVoidMethod(Env, FJavaWrapper::GameActivityThis, Method, jKey, Value);
	}
#endif
}

void AirbridgeAndroidCore::SetUserAttribute(const FString& Key, const bool Value)
{
#if PLATFORM_ANDROID
    if (JNIEnv* Env = FAndroidApplication::GetJavaEnv())
    {
        jmethodID Method = FJavaWrapper::FindMethod(Env,
            FJavaWrapper::GameActivityClassID,
            "airbridgeSetUserAttribute", "(Ljava/lang/String;Z)V", false);
        jstring jKey = Env->NewStringUTF(TCHAR_TO_UTF8(*Key));
        FJavaWrapper::CallVoidMethod(Env, FJavaWrapper::GameActivityThis, Method, jKey, Value);
    }
#endif
}

void AirbridgeAndroidCore::SetUserAttribute(const FString& Key, const FString& Value)
{
#if PLATFORM_ANDROID
    if (JNIEnv* Env = FAndroidApplication::GetJavaEnv())
    {
        jmethodID Method = FJavaWrapper::FindMethod(Env,
            FJavaWrapper::GameActivityClassID,
            "airbridgeSetUserAttribute", "(Ljava/lang/String;Ljava/lang/String;)V", false);
        jstring jKey = Env->NewStringUTF(TCHAR_TO_UTF8(*Key));
        jstring jValue = Env->NewStringUTF(TCHAR_TO_UTF8(*Value));
        FJavaWrapper::CallVoidMethod(Env, FJavaWrapper::GameActivityThis, Method, jKey, jValue);
    }
#endif
}

void AirbridgeAndroidCore::RemoveUserAttribute(const FString& Key)
{
#if PLATFORM_ANDROID
	AirbridgeJNI::GameActivityCallMethod_OneParam_String("airbridgeRemoveUserAttribute", Key);
#endif
}

void AirbridgeAndroidCore::ClearUserAttributes()
{
#if PLATFORM_ANDROID
	AirbridgeJNI::GameActivityCallMethod("airbridgeClearUserAttributes");
#endif
}

void AirbridgeAndroidCore::SetUserAlias(const FString& Key, const FString& Value)
{
#if PLATFORM_ANDROID
    if (JNIEnv* Env = FAndroidApplication::GetJavaEnv())
    {
        jmethodID Method = FJavaWrapper::FindMethod(Env,
            FJavaWrapper::GameActivityClassID,
            "airbridgeSetUserAlias", "(Ljava/lang/String;Ljava/lang/String;)V", false);
        jstring jKey = Env->NewStringUTF(TCHAR_TO_UTF8(*Key));
        jstring jValue = Env->NewStringUTF(TCHAR_TO_UTF8(*Value));
        FJavaWrapper::CallVoidMethod(Env, FJavaWrapper::GameActivityThis, Method, jKey, jValue);
    }
#endif
}

void AirbridgeAndroidCore::RemoveUserAlias(const FString& Key)
{
#if PLATFORM_ANDROID
	AirbridgeJNI::GameActivityCallMethod_OneParam_String("airbridgeRemoveUserAlias", Key);
#endif
}

void AirbridgeAndroidCore::ClearUserAlias()
{
#if PLATFORM_ANDROID
	AirbridgeJNI::GameActivityCallMethod("airbridgeClearUserAlias");
#endif
}

void AirbridgeAndroidCore::ClearUser()
{
#if PLATFORM_ANDROID
	AirbridgeJNI::GameActivityCallMethod("airbridgeClearUser");
#endif
}

void AirbridgeAndroidCore::SetDeviceAlias(const FString& Key, const FString& Value)
{
#if PLATFORM_ANDROID
    if (JNIEnv* Env = FAndroidApplication::GetJavaEnv())
    {
        jmethodID Method = FJavaWrapper::FindMethod(Env,
            FJavaWrapper::GameActivityClassID,
            "airbridgeSetDeviceAlias", "(Ljava/lang/String;Ljava/lang/String;)V", false);
        jstring JStringKey = Env->NewStringUTF(TCHAR_TO_UTF8(*Key));
        jstring JStringValue = Env->NewStringUTF(TCHAR_TO_UTF8(*Value));
        FJavaWrapper::CallVoidMethod(Env, FJavaWrapper::GameActivityThis, Method, JStringKey, JStringValue);
    }
#endif
}

void AirbridgeAndroidCore::RemoveDeviceAlias(const FString& Key)
{
#if PLATFORM_ANDROID
	AirbridgeJNI::GameActivityCallMethod_OneParam_String("airbridgeRemoveDeviceAlias", Key);
#endif
}

void AirbridgeAndroidCore::ClearDeviceAlias()
{
#if PLATFORM_ANDROID
	AirbridgeJNI::GameActivityCallMethod("airbridgeClearDeviceAlias");
#endif
}

void AirbridgeAndroidCore::RegisterPushToken(const FString& Token)
{
#if PLATFORM_ANDROID
	AirbridgeJNI::GameActivityCallMethod_OneParam_String("airbridgeRegisterPushToken", Token);
#endif
}

#pragma endregion

#pragma region Event

void AirbridgeAndroidCore::TrackEvent(
	const FString& Category,
	const UAirbridgeMap* SemanticAttributes,
	const UAirbridgeMap* CustomAttributes
)
{
#if PLATFORM_ANDROID
    if (JNIEnv* Env = FAndroidApplication::GetJavaEnv())
    {
        jmethodID Method = FJavaWrapper::FindMethod(Env,
           FJavaWrapper::GameActivityClassID,
           "airbridgeTrackEvent",
           "(Ljava/lang/String;Ljava/util/HashMap;Ljava/util/HashMap;)V", false);
		// Category
    	jstring JStringCategory = Env->NewStringUTF(TCHAR_TO_UTF8(*Category));
		// Semantic Attributes
    	jobject JSemanticAttributes = NULL;
        if (SemanticAttributes != nullptr)
        {
            JSemanticAttributes = (jobject) SemanticAttributes->CreateJavaObject();
        }
    	// Custom Attributes
    	jobject JCustomAttributes = NULL;
    	if (CustomAttributes != nullptr)
    	{
    		JCustomAttributes = (jobject) CustomAttributes->CreateJavaObject();
    	}
        FJavaWrapper::CallVoidMethod(Env, FJavaWrapper::GameActivityThis, Method,
        	JStringCategory, JSemanticAttributes, JCustomAttributes);
    }
#endif
}

#pragma endregion

#pragma region Placement

void AirbridgeAndroidCore::Click(
	const FString& TrackingLink,
	const AirbridgeVoidParamFunction OnSuccess,
	const AirbridgeStringParamFunction OnFailure
)
{
	FString UUID = FGuid::NewGuid().ToString();
	
	SetAirbridgeVoidParamFunctionMap(Click_OnSuccess, UUID, OnSuccess);
	SetAirbridgeStringParamFunctionMap(Click_OnFailure, UUID, OnFailure);
#if PLATFORM_ANDROID
	AirbridgeJNI::GameActivityCallMethod_TwoParam_String_String("airbridgeClick", TrackingLink, UUID);
#endif
}

void AirbridgeAndroidCore::Impression(
	const FString& TrackingLink,
	const AirbridgeVoidParamFunction OnSuccess,
	const AirbridgeStringParamFunction OnFailure
)
{
	FString UUID = FGuid::NewGuid().ToString();
	
	SetAirbridgeVoidParamFunctionMap(Impression_OnSuccess, UUID, OnSuccess);
	SetAirbridgeStringParamFunctionMap(Impression_OnFailure, UUID, OnFailure);
#if PLATFORM_ANDROID
	AirbridgeJNI::GameActivityCallMethod_TwoParam_String_String("airbridgeImpression", TrackingLink, UUID);
#endif
}

#pragma endregion

#pragma region Publication

void AirbridgeAndroidCore::FetchDeviceUUID(
	const AirbridgeStringParamFunction OnSuccess,
	const AirbridgeStringParamFunction OnFailure,
	bool UseBroadcast
)
{
	FString UUID = FGuid::NewGuid().ToString();
	if (!UseBroadcast)
	{
		SetAirbridgeStringParamFunctionMap(FetchDeviceUUID_OnSuccess, UUID, OnSuccess);
		SetAirbridgeStringParamFunctionMap(FetchDeviceUUID_OnFailure, UUID, OnFailure);
	}
#if PLATFORM_ANDROID
	AirbridgeJNI::GameActivityCallMethod_TwoParam_String_Boolean("airbridgeFetchDeviceUUID", UUID, UseBroadcast);
#endif
}

void AirbridgeAndroidCore::FetchAirbridgeGeneratedUUID(
	const AirbridgeStringParamFunction OnSuccess,
	const AirbridgeStringParamFunction OnFailure,
	bool UseBroadcast
)
{
	FString UUID = FGuid::NewGuid().ToString();
	if (!UseBroadcast)
	{
		SetAirbridgeStringParamFunctionMap(FetchAirbridgeGeneratedUUID_OnSuccess, UUID, OnSuccess);
		SetAirbridgeStringParamFunctionMap(FetchAirbridgeGeneratedUUID_OnFailure, UUID, OnSuccess);
	}
#if PLATFORM_ANDROID
	AirbridgeJNI::GameActivityCallMethod_TwoParam_String_Boolean("airbridgeFetchAirbridgeGeneratedUUID", UUID, UseBroadcast);
#endif
}

void AirbridgeAndroidCore::SetOnAttributionReceived(const AirbridgeMapParamFunction OnAttributionReceived, bool UseBroadcast)
{
	// Only the first callback is valid
	if (DidSetUnrealBroadcastAttribution || AirbridgeOnAttributionResultReceiveListener) return;

	if (UseBroadcast)
	{
		DidSetUnrealBroadcastAttribution = true;
	}
	else
	{
		AirbridgeOnAttributionResultReceiveListener = OnAttributionReceived;	
	}

	HandleOnAttributionReceived();
}

#pragma endregion

#pragma region Hybrid

FString AirbridgeAndroidCore::CreateWebInterfaceScript(const FString& WebToken, const FString& PostMessageScript)
{
#if PLATFORM_ANDROID
	if (JNIEnv* Env = FAndroidApplication::GetJavaEnv())
	{
		jmethodID Method = FJavaWrapper::FindMethod(Env,
			FJavaWrapper::GameActivityClassID,
			"airbridgeCreateWebInterfaceScript",
			"(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;", false);
		jstring JWebToken = Env->NewStringUTF(TCHAR_TO_UTF8(*WebToken));
		jstring JPostMessageScript = Env->NewStringUTF(TCHAR_TO_UTF8(*PostMessageScript));
		jstring JWebInterfaceScript = (jstring) FJavaWrapper::CallObjectMethod(Env, FJavaWrapper::GameActivityThis, Method, JWebToken, JPostMessageScript);
		
		// Copies the new key into its final C string buffer
		const char* CWebInterfaceScript = Env->GetStringUTFChars(JWebInterfaceScript, nullptr);
		if (CWebInterfaceScript == nullptr) {
			return "";  /* OutOfMemoryError already thrown */
		}
	
		FString toReturn = FString(CWebInterfaceScript);

		//release heap memory
		Env->ReleaseStringUTFChars(JWebInterfaceScript, CWebInterfaceScript);
		Env->DeleteLocalRef(JWebInterfaceScript);
		
		return toReturn;
	}
	return "";
#endif
	return "";
}

void AirbridgeAndroidCore::HandleWebInterfaceCommand(const FString& Command)
{
#if PLATFORM_ANDROID
	AirbridgeJNI::GameActivityCallMethod_OneParam_String("airbridgeHandleWebInterfaceCommand", Command);
#endif
}

#pragma endregion


void AirbridgeAndroidCore::ProcessHandleDeeplink(const FString& Uri)
{
	if (DidSetUnrealBroadcastDeeplink || AirbridgeStringParamFunctionMap.Contains(OnDeeplinkReceived))
	{
		HandleDeeplinkOnSuccess(Uri);
	}
	else
	{
		StartDeeplink = Uri;
	}
}

void AirbridgeAndroidCore::ProcessHandleDeferredDeeplink(const FString& Uri)
{
	if (DidSetUnrealBroadcastDeeplink || AirbridgeStringParamFunctionMap.Contains(OnDeeplinkReceived))
	{
		HandleDeeplinkOnSuccess(Uri);
	}
}

void AirbridgeAndroidCore::HandleDeeplinkOnSuccess(const FString& Uri)
{
	if (Uri.IsEmpty()) return;

	if (DidSetUnrealBroadcastDeeplink)
	{
		// Send Deeplink Url to Unreal Engine GameMode Objects
		for (TObjectIterator<UAirbridgeCallbacks> Itr; Itr; ++Itr)
		{
			Itr->OnDeeplinkReceived.Broadcast(Uri);
		}
	}
	else
	{
		AirbridgeStringParamFunction OnDeeplinkReceivedCallback = AirbridgeStringParamFunctionMap[OnDeeplinkReceived][""];
		if (OnDeeplinkReceivedCallback) OnDeeplinkReceivedCallback(Uri);	
	}
	
	StartDeeplink.Reset();
}

void AirbridgeAndroidCore::ProcessClickOnSuccess(const FString& MethodUUID)
{
	// Check if callback method exists
	if (AirbridgeVoidParamFunctionMap.Contains(Click_OnSuccess) &&
		AirbridgeVoidParamFunctionMap[Click_OnSuccess].Contains(MethodUUID))
	{
		// Call callback method
		AirbridgeVoidParamFunction CallbackFunction = AirbridgeVoidParamFunctionMap[Click_OnSuccess][MethodUUID];
		if (CallbackFunction) CallbackFunction();
		// Release callback method
		AirbridgeVoidParamFunctionMap[Click_OnSuccess].Remove(MethodUUID);
	}
}

void AirbridgeAndroidCore::ProcessClickOnFailure(const FString& MethodUUID, const FString& Error)
{
	// Check if callback method exists
	if (AirbridgeStringParamFunctionMap.Contains(Click_OnFailure) &&
		AirbridgeStringParamFunctionMap[Click_OnFailure].Contains(MethodUUID))
	{
		// Call callback method
		AirbridgeStringParamFunction CallbackFunction = AirbridgeStringParamFunctionMap[Click_OnFailure][MethodUUID];
		if (CallbackFunction) CallbackFunction(Error);
		// Release callback method
		AirbridgeStringParamFunctionMap[Click_OnFailure].Remove(MethodUUID);
	}
}

void AirbridgeAndroidCore::ProcessImpressionOnSuccess(const FString& MethodUUID)
{
	// Check if callback method exists
	if (AirbridgeVoidParamFunctionMap.Contains(Impression_OnSuccess) &&
		AirbridgeVoidParamFunctionMap[Impression_OnSuccess].Contains(MethodUUID))
	{
		// Call callback method
		AirbridgeVoidParamFunction CallbackFunction = AirbridgeVoidParamFunctionMap[Impression_OnSuccess][MethodUUID];
		if (CallbackFunction) CallbackFunction();
		// Release callback method
		AirbridgeVoidParamFunctionMap[Impression_OnSuccess].Remove(MethodUUID);
	}
}

void AirbridgeAndroidCore::ProcessImpressionOnFailure(const FString& MethodUUID, const FString& Error)
{
	// Check if callback method exists
	if (AirbridgeStringParamFunctionMap.Contains(Impression_OnFailure) &&
		AirbridgeStringParamFunctionMap[Impression_OnFailure].Contains(MethodUUID))
	{
		// Call callback method
		AirbridgeStringParamFunction CallbackFunction = AirbridgeStringParamFunctionMap[Impression_OnFailure][MethodUUID];
		if (CallbackFunction) CallbackFunction(Error);
		// Release callback method
		AirbridgeStringParamFunctionMap[Impression_OnFailure].Remove(MethodUUID);
	}
}

void AirbridgeAndroidCore::UnrealBroadcastDeviceUUID(const FString& DeviceUUID)
{
	for (TObjectIterator<UAirbridgeCallbacks> Itr; Itr; ++Itr)
	{
		Itr->OnDeviceUUIDFetched.Broadcast(DeviceUUID);
	}
}

void AirbridgeAndroidCore::ProcessFetchDeviceUUIDOnSuccess(const FString& MethodUUID, const FString& DeviceUUID)
{
	// Check if callback method exists
	if (AirbridgeStringParamFunctionMap.Contains(FetchDeviceUUID_OnSuccess) &&
		AirbridgeStringParamFunctionMap[FetchDeviceUUID_OnSuccess].Contains(MethodUUID))
	{
		// Call callback method
		AirbridgeStringParamFunction CallbackFunction = AirbridgeStringParamFunctionMap[FetchDeviceUUID_OnSuccess][MethodUUID];
		if (CallbackFunction) CallbackFunction(DeviceUUID);
		// Release callback method
		AirbridgeStringParamFunctionMap[FetchDeviceUUID_OnSuccess].Remove(MethodUUID);
	}
}

void AirbridgeAndroidCore::ProcessFetchDeviceUUIDOnFailure(const FString& MethodUUID, const FString& Error)
{
	// Check if callback method exists
	if (AirbridgeStringParamFunctionMap.Contains(FetchDeviceUUID_OnFailure) &&
		AirbridgeStringParamFunctionMap[FetchDeviceUUID_OnFailure].Contains(MethodUUID))
	{
		// Call callback method
		AirbridgeStringParamFunction CallbackFunction = AirbridgeStringParamFunctionMap[FetchDeviceUUID_OnFailure][MethodUUID];
		if (CallbackFunction) CallbackFunction(Error);
		// Release callback method
		AirbridgeStringParamFunctionMap[FetchDeviceUUID_OnFailure].Remove(MethodUUID);
	}
}

void AirbridgeAndroidCore::UnrealBroadcastAirbridgeGeneratedUUID(const FString& AirbridgeGeneratedUUID)
{
	for (TObjectIterator<UAirbridgeCallbacks> Itr; Itr; ++Itr)
	{
		Itr->OnAirbridgeGeneratedUUIDFetched.Broadcast(AirbridgeGeneratedUUID);
	}
}

void AirbridgeAndroidCore::ProcessFetchAirbridgeGeneratedUUIDOnSuccess(const FString& MethodUUID, const FString& AirbridgeGeneratedUUID)
{
	// Check if callback method exists
	if (AirbridgeStringParamFunctionMap.Contains(FetchAirbridgeGeneratedUUID_OnSuccess) &&
		AirbridgeStringParamFunctionMap[FetchAirbridgeGeneratedUUID_OnSuccess].Contains(MethodUUID))
	{
		// Call callback method
		AirbridgeStringParamFunction CallbackFunction = AirbridgeStringParamFunctionMap[FetchAirbridgeGeneratedUUID_OnSuccess][MethodUUID];
		if (CallbackFunction) CallbackFunction(AirbridgeGeneratedUUID);
		// Release callback method
		AirbridgeStringParamFunctionMap[FetchAirbridgeGeneratedUUID_OnSuccess].Remove(MethodUUID);
	}
}

void AirbridgeAndroidCore::ProcessFetchAirbridgeGeneratedUUIDOnFailure(const FString& MethodUUID, const FString& Error)
{
	// Check if callback method exists
	if (AirbridgeStringParamFunctionMap.Contains(FetchAirbridgeGeneratedUUID_OnFailure) &&
		AirbridgeStringParamFunctionMap[FetchAirbridgeGeneratedUUID_OnFailure].Contains(MethodUUID))
	{
		// Call callback method
		AirbridgeStringParamFunction CallbackFunction = AirbridgeStringParamFunctionMap[FetchAirbridgeGeneratedUUID_OnFailure][MethodUUID];
		if (CallbackFunction) CallbackFunction(Error);
		// Release callback method
		AirbridgeStringParamFunctionMap[FetchAirbridgeGeneratedUUID_OnFailure].Remove(MethodUUID);
	}
}

void AirbridgeAndroidCore::ProcessOnAttributionResultReceived(const TMap<FString, FString>& Data)
{
	ReceivedAttribution = Data;
	HandleOnAttributionReceived();
}

void AirbridgeAndroidCore::HandleOnAttributionReceived()
{
	if (ReceivedAttribution.Num() == 0) return;

	if (DidSetUnrealBroadcastAttribution)
	{
		// Send Attribution Data to Unreal Engine GameMode Objects
		for (TObjectIterator<UAirbridgeCallbacks> Itr; Itr; ++Itr)
		{
			Itr->OnAttributionReceived.Broadcast(FAirbridgeAttribution{ReceivedAttribution});
		}
	}
	else if (AirbridgeOnAttributionResultReceiveListener)
	{
		AirbridgeOnAttributionResultReceiveListener(ReceivedAttribution);
	}
	else
	{
		return;
	}
	
	ReceivedAttribution.Empty();
}