package co.ab180.airbridge.ue;

import android.app.Activity;
import android.app.Application;
import android.content.ContentProvider;
import android.content.ContentValues;
import android.database.Cursor;
import android.net.Uri;
import android.util.Log;

import co.ab180.airbridge.AirbridgeLifecycleIntegration;
import co.ab180.airbridge.OnAttributionResultReceiveListener;

import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import co.ab180.airbridge.Airbridge;
import co.ab180.airbridge.AirbridgeLogLevel;
import co.ab180.airbridge.AirbridgeOptionBuilder;

public class AirbridgeContentProvider extends ContentProvider {

    private static final String UNREAL_ENGINE_4_GAME_ACTIVITY = "com.epicgames.ue4.GameActivity";
    private static final String UNREAL_ENGINE_5_GAME_ACTIVITY = "com.epicgames.unreal.GameActivity";

    @Override
    public boolean onCreate() {
        try {
            Application app = (Application) Objects.requireNonNull(getContext()).getApplicationContext();

            AirbridgeOptionBuilder builder = new AirbridgeOptionBuilder(AirbridgeSettings.appName, AirbridgeSettings.appToken);
            builder.setSDKDevelopmentPlatform("unreal");

            if (isNotEmpty(AirbridgeSettings.sdkSignatureSecretID) && isNotEmpty(AirbridgeSettings.sdkSignatureSecret)) {
                builder.setSDKSignature(AirbridgeSettings.sdkSignatureSecretID, AirbridgeSettings.sdkSignatureSecret);
            }

            builder.setLogLevel(AirbridgeLogLevel.values()[AirbridgeSettings.logLevel]);

            String customDomain = AirbridgeSettings.customDomain;
            if (isNotEmpty(customDomain)) {
                builder.setCustomDomains(joinedStringToList(customDomain, " "));
            }

            builder
                    .setSessionTimeout(AirbridgeSettings.sessionTimeoutSeconds)
                    .setHashUserInformationEnabled(AirbridgeSettings.userInfoHashEnabled)
                    .setCollectLocationEnabled(AirbridgeSettings.locationCollectionEnabled)
                    .setTrackAirbridgeDeeplinkOnlyEnabled(AirbridgeSettings.trackAirbridgeLinkOnly)
                    .setAutoStartTrackingEnabled(AirbridgeSettings.autoStartTrackingEnabled)
                    .setTrackMetaDeferredAppLinkEnabled(AirbridgeSettings.facebookDeferredAppLinkEnabled);

            builder.setOnAttributionReceived(new OnAttributionResultReceiveListener() {
                @Override
                public void onAttributionResultReceived(@NotNull Map<String, String> map) {
                    if (
                        processSetOnAttributionReceived(UNREAL_ENGINE_4_GAME_ACTIVITY, map) ||
                        processSetOnAttributionReceived(UNREAL_ENGINE_5_GAME_ACTIVITY, map)
                    ) {
                        return;
                    }

                    Log.e("Airbridge Unreal Engine", "Couldn't handle {setOnAttributionReceived} - failed to find GameActivity#airbridgeProcessSetOnAttributionReceived");
                }
            });

            builder
                    .setTrackInSessionLifeCycleEventEnabled(AirbridgeSettings.trackInSessionLifeCycleEventEnabled)
                    .setPauseEventTransmitOnBackgroundEnabled(AirbridgeSettings.pauseEventTransmitOnBackgroundEnabled)
                    .setClearEventBufferOnInitializeEnabled(AirbridgeSettings.clearEventBufferOnInitializeEnabled)
                    .setSDKEnabled(AirbridgeSettings.sdkEnabled);

            String appMarketIdentifier = AirbridgeSettings.appMarketIdentifier;
            if (isNotEmpty(appMarketIdentifier)) {
                builder.setAppMarketIdentifier(appMarketIdentifier);
            }

            builder
                    .setEventBufferCountLimit(AirbridgeSettings.eventBufferCountLimitInGibibyte)
                    .setEventBufferSizeLimit(AirbridgeSettings.eventBufferSizeLimitInGibibyte)
                    .setEventTransmitInterval(AirbridgeSettings.eventTransmitIntervalSeconds);

            String facebookAppId = AirbridgeSettings.facebookAppId;
            if (isNotEmpty(facebookAppId)) {
                builder.setMetaInstallReferrer(facebookAppId);
            }

            builder.setLifecycleIntegration(new AirbridgeLifecycleIntegration() {

                @Nullable
                @Override
                public String getDataString(@NotNull Activity activity) {
                    if (AirbridgeUnreal.airbridgeLifecycleIntegration != null) {
                        return AirbridgeUnreal.airbridgeLifecycleIntegration.getDataString(activity);
                    }
                    return null;
                }
            });

            Airbridge.initializeSDK(app, builder.build());
        } catch (Throwable throwable) {
            Log.e("Airbridge Unreal Engine", "Couldn't initialize SDK", throwable);
        }
        return true;
    }

    @Override
    public Cursor query(@NotNull Uri uri, String[] projection, String selection, String[] selectionArgs, String sortOrder) {
        return null;
    }

    @Override
    public String getType(@NotNull Uri uri) {
        return null;
    }

    @Override
    public Uri insert(@NotNull Uri uri, ContentValues values) {
        return null;
    }

    @Override
    public int delete(@NotNull Uri uri, String selection, String[] selectionArgs) {
        return 0;
    }

    @Override
    public int update(@NotNull Uri uri, ContentValues values, String selection, String[] selectionArgs) {
        return 0;
    }

    private List<String> joinedStringToList(String joinedString, String delimiter) {
        String[] array = joinedString.split(delimiter);
        List<String> ret = new ArrayList<>();
        for (String element : array) {
            // Remove white space
            if (!element.trim().isEmpty()) {
                ret.add(element);
            }
        }
        return ret;
    }

    private boolean processSetOnAttributionReceived(String className, @NotNull Map<String, String> map) {
        try {
            Class<?> clazz = Class.forName(className);
            Method method = clazz.getMethod("airbridgeProcessSetOnAttributionReceived", Map.class);
            method.invoke(null, map);
            return true;
        } catch (Throwable ignore) {
            return false;
        }
    }

    private boolean isNotEmpty(String string) {
        if (string == null) { return false; }
        return !(string.isEmpty());
    }
}