// Copyright © 2023 ab180. All rights reserved.

#include "AirbridgeIOSCore.h"
#include "AirbridgeCallbacks.h"
#include "UObject/UObjectIterator.h"

#include "IOSAppDelegate.h"

#import "Platforms/iOS/Native/AUAppDelegate.h"
#import "Platforms/iOS/Native/AUAppSetting.h"
#import "Platforms/iOS/Native/Utils/AUHex.h"

#import <UIKit/UIKit.h>
#import <Airbridge/Airbridge.h>

FString AirbridgeIOSCore::StartDeeplink;
TMap<FString, FString> AirbridgeIOSCore::ReceivedAttribution;

AirbridgeIOSCore::AirbridgeIOSCore()
{
    [AUAppDelegate load];
}

AirbridgeIOSCore::~AirbridgeIOSCore()
{
    [AUAppDelegate unregisterLifeCycleListener];
}

#pragma region Core

void AirbridgeIOSCore::EnableSDK()
{
	[Airbridge enableSDK];
}

void AirbridgeIOSCore::DisableSDK()
{
	[Airbridge disableSDK];
}

bool AirbridgeIOSCore::IsSDKEnabled()
{
	return Airbridge.isSDKEnabled;
}

#pragma endregion

#pragma region Privacy

void AirbridgeIOSCore::StartTracking()
{
	[Airbridge startTracking];
}

void AirbridgeIOSCore::StopTracking()
{
    [Airbridge stopTracking];
}

bool AirbridgeIOSCore::IsTrackingEnabled()
{
    return Airbridge.isTrackingEnabled;
}

#pragma endregion

#pragma region Deeplink

void AirbridgeIOSCore::SetOnDeeplinkReceived(const AirbridgeStringParamFunction OnDeeplinkReceived, bool UseBroadcast)
{
	if (DidSetUnrealBroadcastDeeplink || AirbridgeHandleDeeplinkFunction) { return; }

	if (UseBroadcast)
	{
		DidSetUnrealBroadcastDeeplink = true;
	}
	else
	{
		AirbridgeHandleDeeplinkFunction = OnDeeplinkReceived;
	}
	
	if (!StartDeeplink.IsEmpty())
	{
		HandleDeeplinkOnSuccess(StartDeeplink);
	}
	else
	{
		[Airbridge handleDeferredDeeplinkOnSuccess:^(NSURL* url) {
			if (url != nil) {
				NSString *urlString = [url absoluteString];
				ProcessHandleDeferredDeeplink(FString(urlString));
			}
		}];
	}
}

#pragma endregion

#pragma region Data Collection

void AirbridgeIOSCore::SetUserID(const FString& Id)
{
	[Airbridge setUserID:Id.GetNSString()];
}

void AirbridgeIOSCore::ClearUserID()
{
	[Airbridge clearUserID];
}

void AirbridgeIOSCore::SetUserEmail(const FString& Email)
{
	[Airbridge setUserEmail:Email.GetNSString()];
}

void AirbridgeIOSCore::ClearUserEmail()
{
	[Airbridge clearUserEmail];
}

void AirbridgeIOSCore::SetUserPhone(const FString& Phone)
{
	[Airbridge setUserPhone:Phone.GetNSString()];
}

void AirbridgeIOSCore::ClearUserPhone()
{
	[Airbridge clearUserPhone];
}

void AirbridgeIOSCore::SetUserAttribute(const FString& Key, const int32 Value)
{
	[Airbridge setUserAttributeWithKey:Key.GetNSString() value:@(Value)];
}

void AirbridgeIOSCore::SetUserAttribute(const FString& Key, const int64 Value)
{
	[Airbridge setUserAttributeWithKey:Key.GetNSString() value:@(Value)];
}

void AirbridgeIOSCore::SetUserAttribute(const FString& Key, const float Value)
{
    [Airbridge setUserAttributeWithKey:Key.GetNSString() value:@(Value)];
}

void AirbridgeIOSCore::SetUserAttribute(const FString& Key, const double Value)
{
    [Airbridge setUserAttributeWithKey:Key.GetNSString() value:@(Value)];
}

void AirbridgeIOSCore::SetUserAttribute(const FString& Key, const bool Value)
{
    [Airbridge setUserAttributeWithKey:Key.GetNSString() value:@(Value)];
}

void AirbridgeIOSCore::SetUserAttribute(const FString& Key, const FString& Value)
{
	[Airbridge setUserAttributeWithKey:Key.GetNSString() value:Value.GetNSString()];
}

void AirbridgeIOSCore::RemoveUserAttribute(const FString& Key)
{
    [Airbridge removeUserAttributeWithKey:Key.GetNSString()];
}

void AirbridgeIOSCore::ClearUserAttributes()
{
	[Airbridge clearUserAttributes];
}

void AirbridgeIOSCore::SetUserAlias(const FString& Key, const FString& Value)
{
	[Airbridge setUserAliasWithKey:Key.GetNSString() value:Value.GetNSString()];
}

void AirbridgeIOSCore::RemoveUserAlias(const FString& Key)
{
	[Airbridge removeUserAliasWithKey:Key.GetNSString()];
}

void AirbridgeIOSCore::ClearUserAlias()
{
	[Airbridge clearUserAlias];
}

void AirbridgeIOSCore::ClearUser()
{
	[Airbridge clearUser];
}

void AirbridgeIOSCore::SetDeviceAlias(const FString& Key, const FString& Value)
{
	[Airbridge setDeviceAliasWithKey:Key.GetNSString() value:Value.GetNSString()];
}

void AirbridgeIOSCore::RemoveDeviceAlias(const FString& Key)
{
    [Airbridge removeDeviceAliasWithKey:Key.GetNSString()];
}

void AirbridgeIOSCore::ClearDeviceAlias()
{
	[Airbridge clearDeviceAlias];
}

void AirbridgeIOSCore::RegisterPushToken(const FString& Token)
{
	NSString *tokenString = Token.GetNSString();
	NSData *tokenData = [AUHex dataFromHexString:tokenString];
	[Airbridge registerPushToken:tokenData];
}

#pragma endregion

#pragma region Event

void AirbridgeIOSCore::TrackEvent(const FString& Category, const UAirbridgeMap* SemanticAttributes, const UAirbridgeMap* CustomAttributes)
{
	[Airbridge
		trackEventWithCategory:Category.GetNSString()
		semanticAttributes:(SemanticAttributes != nullptr) ? SemanticAttributes->ConvertToNSDictionary() : nil
		customAttributes: (CustomAttributes != nullptr) ? CustomAttributes->ConvertToNSDictionary() : nil
	];
}

#pragma endregion

#pragma region Placement

void AirbridgeIOSCore::Click(const FString& TrackingLink, const AirbridgeVoidParamFunction OnSuccess, const AirbridgeStringParamFunction OnFailure)
{
	NSURL *url = [NSURL URLWithString:TrackingLink.GetNSString()];
	
	[Airbridge
		clickWithTrackingLink:url
		onSuccess:^{
			if (OnSuccess) { OnSuccess(); }
		}
		onFailure:^(NSError * _Nonnull error) {
			if (OnFailure) { OnFailure(error.localizedDescription); }
		}];
}

void AirbridgeIOSCore::Impression(const FString& TrackingLink, const AirbridgeVoidParamFunction OnSuccess, const AirbridgeStringParamFunction OnFailure)
{
	NSURL *url = [NSURL URLWithString:TrackingLink.GetNSString()];
    
	[Airbridge
		impressionWithTrackingLink:url
		onSuccess:^{
			if (OnSuccess) { OnSuccess(); }
		}
		onFailure:^(NSError * _Nonnull error) {
			if (OnFailure) { OnFailure(error.localizedDescription); }
		}];
}

#pragma endregion

#pragma region Publication

void AirbridgeIOSCore::FetchDeviceUUID(
	const AirbridgeStringParamFunction OnSuccess,
	const AirbridgeStringParamFunction OnFailure,
	bool UseBroadcast
)
{
	[Airbridge
		fetchDeviceUUIDOnSuccess:^(NSString * _Nonnull deviceUUID) {
			if (OnSuccess) {
				const char* cString = [deviceUUID UTF8String];
				OnSuccess(FString(UTF8_TO_TCHAR(cString)));
			}
		}
		onFailure:^(NSError * _Nonnull error) {
			if (OnFailure) {
				const char* cString = [error.localizedDescription UTF8String];
				OnFailure(FString(UTF8_TO_TCHAR(cString)));
			}
		}];
}

void AirbridgeIOSCore::FetchAirbridgeGeneratedUUID(
	const AirbridgeStringParamFunction OnSuccess,
	const AirbridgeStringParamFunction OnFailure,
	bool UseBroadcast
)
{
	[Airbridge
		fetchAirbridgeGeneratedUUIDOnSuccess:^(NSString * _Nonnull airbridgeGeneratedUUID) {
			if (OnSuccess) {
				const char* cString = [airbridgeGeneratedUUID UTF8String];
				OnSuccess(FString(UTF8_TO_TCHAR(cString)));
			}
		}
		onFailure:^(NSError * _Nonnull error) {
			if (OnFailure) {
				const char* cString = [error.localizedDescription UTF8String];
				OnFailure(FString(UTF8_TO_TCHAR(cString)));
			}
		}];
}

void AirbridgeIOSCore::SetOnAttributionReceived(const AirbridgeMapParamFunction OnAttributionReceived, bool UseBroadcast)
{
	if (DidSetUnrealBroadcastAttribution || AirbridgeOnAttributionResultReceiveListener) return;

	if (UseBroadcast)
	{
		DidSetUnrealBroadcastAttribution = true;
	}
	else
	{
		AirbridgeOnAttributionResultReceiveListener = OnAttributionReceived;	
	}

	HandleOnAttributionReceived();
}

#pragma endregion

#pragma region Hybrid

FString AirbridgeIOSCore::CreateWebInterfaceScript(const FString& WebToken, const FString& PostMessageScript)
{
    NSString *script = [Airbridge
        createWebInterfaceScriptWithWebToken:WebToken.GetNSString()
        postMessageScript:PostMessageScript.GetNSString()
    ];
    
    const char* cString = [script UTF8String];
    return FString(cString);
}

void AirbridgeIOSCore::HandleWebInterfaceCommand(const FString& Command)
{
    [Airbridge handleWebInterfaceCommand:Command.GetNSString()];
}

#pragma endregion

void AirbridgeIOSCore::ProcessHandleDeeplink(const FString& Uri)
{
	if (DidSetUnrealBroadcastDeeplink || AirbridgeHandleDeeplinkFunction)
	{
		HandleDeeplinkOnSuccess(Uri);
	}
	else
	{
		StartDeeplink = Uri;
	}
}

void AirbridgeIOSCore::ProcessHandleDeferredDeeplink(const FString& Uri)
{
	if (DidSetUnrealBroadcastDeeplink || AirbridgeHandleDeeplinkFunction)
	{
		HandleDeeplinkOnSuccess(Uri);
	}
}

void AirbridgeIOSCore::HandleDeeplinkOnSuccess(const FString& Uri)
{
	if (Uri.IsEmpty()) return;

	if (DidSetUnrealBroadcastDeeplink)
	{
		// Send Deeplink Url to Unreal Engine GameMode Objects
		for (TObjectIterator<UAirbridgeCallbacks> Itr; Itr; ++Itr)
		{
			Itr->OnDeeplinkReceived.Broadcast(Uri);
		}
	}
	else
	{
		if (AirbridgeHandleDeeplinkFunction) AirbridgeHandleDeeplinkFunction(Uri);	
	}
	
	StartDeeplink.Reset();
}

void AirbridgeIOSCore::ProcessOnAttributionResultReceived(const TMap<FString, FString>& Data)
{
	ReceivedAttribution = Data;
	HandleOnAttributionReceived();
}

void AirbridgeIOSCore::HandleOnAttributionReceived()
{
	if (ReceivedAttribution.Num() == 0) return;

	if (DidSetUnrealBroadcastAttribution)
	{
		// Send Attribution Data to Unreal Engine GameMode Objects
		for (TObjectIterator<UAirbridgeCallbacks> Itr; Itr; ++Itr)
		{
			Itr->OnAttributionReceived.Broadcast(FAirbridgeAttribution{ReceivedAttribution});
		}
	}
	else if (AirbridgeOnAttributionResultReceiveListener)
	{
		AirbridgeOnAttributionResultReceiveListener(ReceivedAttribution);
	}
	else
	{
		return;
	}
	
	ReceivedAttribution.Empty();
}

//MARK: - IOSAppDelegate+AirbridgeAppDelegate
@interface IOSAppDelegate (AirbridgeAppDelegate)

@end

@implementation IOSAppDelegate (AirbridgeAppDelegate)

+ (void)processDeeplinkURL:(NSURL * __nullable)url {
    if (!url) { return; }
    
    [self postDeeplinkCallback:url.absoluteString];
}

+ (void)postDeeplinkCallback:(NSString *)deeplinkString {
    FString string(deeplinkString);
    
    AsyncTask(ENamedThreads::GameThread, [string = MoveTemp(string)]() mutable
    {
        if (IOS_CORE_POINTER) { IOS_CORE_POINTER->ProcessHandleDeeplink(string); }
    });
}


- (BOOL)application:(UIApplication *)application openURL:(NSURL *)url options:(NSDictionary<UIApplicationOpenURLOptionsKey, id>*)options
{
	[Airbridge trackDeeplinkWithUrl:url];
    
    BOOL isHandled = [Airbridge handleDeeplinkWithUrl:url onSuccess:^(NSURL* url) {
        if (!GIsRunning) {
            AirbridgeIOSCore::StartDeeplink = FString(url.absoluteString);
        }
        else {
            [IOSAppDelegate processDeeplinkURL:url];
        }
	}];
    
    if (isHandled || AUAppSetting.isHandleAirbridgeDeeplinkOnly) { return YES; }
    if (!GIsRunning) {
        AirbridgeIOSCore::StartDeeplink = FString(url.absoluteString);
    } else {
        [IOSAppDelegate processDeeplinkURL:url];
    }
    
	return YES;
}

- (BOOL)application:(UIApplication*)application continueUserActivity:(NSUserActivity*)userActivity restorationHandler:(void (^)(NSArray* _Nullable))restorationHandler
{
	[Airbridge trackDeeplinkWithUserActivity:userActivity];
    NSURL *url = userActivity.webpageURL;
    if (!url) { return YES; }
    
    BOOL isHandled = [Airbridge handleDeeplinkWithUserActivity:userActivity onSuccess:^(NSURL* url) {
        if (!GIsRunning) {
            AirbridgeIOSCore::StartDeeplink = FString(url.absoluteString);
        }
        else {
            [IOSAppDelegate processDeeplinkURL:url];
        }
	}];
    
    if (isHandled || AUAppSetting.isHandleAirbridgeDeeplinkOnly) { return YES; }
    if (!GIsRunning) {
        AirbridgeIOSCore::StartDeeplink = FString(url.absoluteString);
    } else {
        [IOSAppDelegate processDeeplinkURL:url];
    }
	
	return YES;
}

@end
